@@ -1,5 +1,5 @@
 /*
-Copyright (c) 2014, 2019, Felix J. Ogris
+Copyright (c) 2014, 2019, 2021, Felix J. Ogris
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
@@ -46,6 +46,9 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 #include <netinet/ip.h>
 #include <netinet/ip6.h>
 #include <netinet/tcp.h>
+#ifdef USE_NG
+#  include <netgraph.h>
+#endif
 #include <capsicum_helpers.h>
 #include <libcasper.h>
 #include <casper/cap_syslog.h>
@@ -267,9 +270,14 @@ void parse_v6(void *buf, unsigned int len, struct rule *rule)
 void print_help(FILE *f)
 {
   fprintf(f,
-"ftp6proxy V0.3 - IPv6/IPv4 FTP firewall helper\n"
+"ftp6proxy V0.4 - IPv6/IPv4 FTP firewall helper\n"
 "\n"
-"compiled-in firewall module: %s\n"
+"compiled-in firewall module: %s"
+#ifdef USE_NG
+"/netgraph\n"
+#else
+"\n"
+#endif
 "\n"
 "usage: ftp6proxy [-b <base>] [-c <count>] [-d [-d]] [-f <facility>]\n"
 "                 [-i <pidfile>] [-j <target>] [-p <port>] [-q <queue>]\n"
@@ -286,7 +294,11 @@ void print_help(FILE *f)
 "  -i <pidfile>   write pid to <pidfile> instead of /var/run/ftp6proxy.pid\n"
 "  -j <target>    call this ipfw rule number instead of allowing the data\n"
 "                 connection\n"
+#ifdef USE_NG
+"  -p <hook>      ng_ipfw(4) hook to connect to; defaults to 10004\n"
+#else
 "  -p <port>      divert(4) port number to listen on; defaults to 10004\n"
+#endif
 "  -q <queue>     place data connections into this pf QoS class\n"
 "  -r <dir>       chroot into <dir> instead of /var/empty\n"
 "  -t <timeout>   wait up to <timeout> seconds for data connections;\n"
@@ -318,17 +330,26 @@ int main(int argc, char **argv)
   gid_t gid;
   uid_t uid;
   struct passwd *pw;
+  char pidstr[16];
   pid_t pid;
   struct stat st;
   FILE *f;
-  char buf[65536];
   struct sigaction sa;
-  struct sockaddr_in sin;
   int port = 10004;
-  int sock;
+  unsigned char buf[8192];
   ssize_t recvlen;
+#ifdef USE_NG
+  int cs, ds;
+#else
+  int kq;
+  struct kevent kev;
+  struct timespec ktimeout;
+  int numkevents;
+  struct sockaddr_in sin;
+  int sock;
   ssize_t sentlen;
   socklen_t sinsize;
+#endif
   struct ip *ip;
   time_t *rules; /* ring buffer */
   int rule_base = 64000;
@@ -338,10 +359,6 @@ int main(int argc, char **argv)
   struct rule rule;
   int timeout = 2;
   void *fw;
-  int kq;
-  struct kevent kev;
-  struct timespec ktimeout;
-  int numkevents;
   cap_rights_t rights;
   cap_channel_t *capcas;
 
@@ -415,12 +432,22 @@ int main(int argc, char **argv)
     if (lstat(pidfile, &st) == 0) {
       if (!S_ISREG(st.st_mode))
         ERRDIEX("%s exists, but is not a regular file", pidfile);
-      if ((f = fopen(pidfile, "r")) == NULL) ERRDIE("fopen(%s, r)", pidfile);
-      if (fgets(buf, sizeof(buf), f) == NULL) ERRDIE("fgets(%s)", pidfile);
-      pid = atoi(buf);
+
+      if ((f = fopen(pidfile, "r")) == NULL)
+        ERRDIE("fopen(%s, r)", pidfile);
+
+      if (fgets(pidstr, sizeof(pidstr), f) == NULL)
+        ERRDIE("fgets(%s)", pidfile);
+
+      pid = atoi(pidstr);
       fclose(f);
-      if (kill(pid, 0) == 0) ERRDIEX("still running (pid=%li)", (long)pid);
-    } else if (errno != ENOENT) ERRDIE("%s", pidfile);
+
+      if (kill(pid, 0) == 0)
+        ERRDIEX("still running (pid=%li)", (long)pid);
+    } else {
+      if (errno != ENOENT)
+        ERRDIE("%s", pidfile);
+    }
 
     /* daemonize */
     pid = fork();
@@ -437,6 +464,33 @@ int main(int argc, char **argv)
   fw = fw_init();
   if (fw == NULL) ERRDIE("cannot initialize firewall");
 
+  /* allocate ring buffer */
+  rules = malloc(rule_count * sizeof(time_t));
+  if (rules == NULL) ERRDIE("cannot allocate rules buffer");
+
+  /* set up signal handlers */
+  memset(&sa, 0, sizeof(sa));
+  sa.sa_handler = kill_handler;
+  if (sigaction(SIGTERM, &sa, NULL) < 0)
+    ERRDIE("sigaction(SIGTERM)");
+
+#ifdef USE_NG
+  /* create ng_socket node */
+  if (NgMkSockNode("ftp6proxy", &cs, &ds) == -1)
+    ERRDIE("NgMkSockNode(ftp6proxy)");
+
+  if (NgSendAsciiMsg(cs, ".:", "connect { path=\"ipfw:\" ourhook=\"fromfw\" "
+                               "peerhook=\"%i\" }", port) == -1)
+    ERRDIE("NgSendAsciiMsg(connect ipfw fromipfw %i)", port);
+
+  cap_rights_init(&rights);
+  if (caph_rights_limit(cs, &rights) != 0)
+    ERRDIE("cannot caph_rights_limit() on netgraph control socket");
+
+  cap_rights_init(&rights, CAP_RECV);
+  if (caph_rights_limit(ds, &rights) != 0)
+    ERRDIE("cannot caph_rights_limit(RECV) on netgraph data socket");
+#else
   /* create divert socket */
   if ((sock = socket(PF_INET, SOCK_RAW, IPPROTO_DIVERT)) < 0)
     ERRDIE("socket()");
@@ -449,10 +503,6 @@ int main(int argc, char **argv)
   if (bind(sock, (struct sockaddr*) &sin, sizeof(struct sockaddr_in)) < 0)
     ERRDIE("bind()");
 
-  /* allocate ring buffer */
-  rules = malloc(rule_count * sizeof(time_t));
-  if (rules == NULL) ERRDIE("cannot allocate rules buffer");
-
   /* we can't block on the socket in order to remove temporary rules */
   kq = kqueue();
   if (kq == -1) ERRDIE("kqueue()");
@@ -461,12 +511,6 @@ int main(int argc, char **argv)
   if (kevent(kq, &kev, 1, NULL, 0, NULL) == -1)
     ERRDIE("cannot add divert socket to kqueue");
 
-  /* set up signal handlers */
-  memset(&sa, 0, sizeof(sa));
-  sa.sa_handler = kill_handler;
-  if (sigaction(SIGTERM, &sa, NULL) < 0)
-    ERRDIE("sigaction(SIGTERM)");
-
   /* limit kqueue and divert socket via capsicum */
   cap_rights_init(&rights, CAP_KQUEUE_EVENT);
   if (caph_rights_limit(kq, &rights) != 0)
@@ -476,6 +520,7 @@ int main(int argc, char **argv)
   if (caph_rights_limit(sock, &rights) != 0)
     ERRDIE("cannot caph_rights_limit(EVENT,RECV,SEND,CONNECT) on divert "
                                                                     "socket");
+#endif
 
   if (debug <= 1) {
     if (chdir("/") != 0) ERRDIE("chdir(/)");
@@ -504,11 +549,11 @@ int main(int argc, char **argv)
       ERRDIE("cannot caph_rights_limit(WRITE) on stderr");
   }
 
+#ifdef USE_NG
   /* limit via capsicum */
-/* XXX not yet possible due to sendto() with non-NULL destination XXX
   if (caph_enter() != 0)
     ERRDIE("caph_enter()");
-*/
+#endif
 
   LOGINFOX("starting...");
 
@@ -525,6 +570,23 @@ int main(int argc, char **argv)
       if (rule_start >= rule_count) rule_start = 0;
     }
 
+#ifdef USE_NG
+    recvlen = NgRecvData(ds, buf, sizeof(buf), NULL);
+    LOGDEBUGX("recvlen=%zi", recvlen);
+
+    if (recvlen == 0) {
+      LOGWARNX("NgRecvData(): data socket closed");
+      running = 0;
+      continue;
+    }
+
+    if (recvlen < 0) {
+      /* NgRecvData() failed */
+      if (running) LOGERR("NgRecvData()");
+      if (errno != EINTR) running = 0;
+      continue;
+    }
+#else
     ktimeout.tv_sec = 1;
     ktimeout.tv_nsec = 0;
     numkevents = kevent(kq, NULL, 0, &kev, 1, &ktimeout);
@@ -540,6 +602,7 @@ int main(int argc, char **argv)
     sinsize = sizeof(sin);
     recvlen = recvfrom(sock, buf, sizeof(buf), 0, (struct sockaddr*) &sin,
                        &sinsize);
+
     LOGDEBUGX("recvlen=%zi", recvlen);
 
     if (recvlen < 0) {
@@ -548,6 +611,7 @@ int main(int argc, char **argv)
       if (errno != EINTR) running = 0;
       continue;
     }
+#endif
 
     rule.action = NONE;
 
@@ -583,11 +647,13 @@ int main(int argc, char **argv)
       }
     }
 
+#ifndef USE_NG
     /* flush buffer */
     sentlen = sendto(sock, buf, recvlen, 0, (struct sockaddr*) &sin, sinsize);
     LOGDEBUGX("sentlen=%zi", sentlen);
     if ((sentlen >= 0) && (sentlen != recvlen))
       LOGWARNX("sent %zi of %zi bytes", sentlen, recvlen);
+#endif
   }
 
   while (rule_start != rule_end) {
