@@ -10,6 +10,7 @@
 #include <netinet/tcp.h>
 #include <net/if.h>
 #include <net/pfvar.h>
+#include <capsicum_helpers.h>
 #include "fw.h"
 
 #define PF_ANCHOR_NAME "ftp6proxy"
@@ -22,17 +23,46 @@ struct pf_data {
   struct pfioc_rule pfr;
 };
 
+const char *fw_ident ()
+{
+  return "pf";
+}
+
 void *fw_init ()
 {
+  int save_errno;
   struct pf_data *pfd;
+  cap_rights_t rights;
+  unsigned long allowed_ioctls[] = { DIOCXBEGIN, DIOCXCOMMIT, DIOCXROLLBACK,
+                                     DIOCADDRULE, DIOCBEGINADDRS };
 
   pfd = malloc(sizeof(*pfd));
-  if (pfd == NULL) return NULL;
+  if (pfd == NULL) goto FW_INIT_ERR0;
 
   pfd->fd = open("/dev/pf", O_RDWR);
-  if (pfd->fd != -1) return pfd;
+  if (pfd->fd < 0) goto FW_INIT_ERR1;
 
+  cap_rights_init(&rights, CAP_IOCTL);
+  if (caph_rights_limit(pfd->fd, &rights) != 0)
+    goto FW_INIT_ERR2;
+
+  if (caph_ioctls_limit(pfd->fd, allowed_ioctls,
+                        sizeof(allowed_ioctls)/sizeof(allowed_ioctls[0])) != 0)
+    goto FW_INIT_ERR2;
+
+  return pfd;
+
+FW_INIT_ERR2:
+  save_errno = errno;
+  close(pfd->fd);
+  errno = save_errno;
+
+FW_INIT_ERR1:
+  save_errno = errno;
   free(pfd);
+  errno = save_errno;
+
+FW_INIT_ERR0:
   return NULL;
 }
 
@@ -45,7 +75,8 @@ int fw_close (void *data)
   return (close(fd) == -1 ? -1 : 0);
 }
 
-static int pf_prepare (struct pf_data *pfd, uint32_t rulenum)
+static int pf_prepare (struct pf_data *pfd, uint32_t rulenum,
+                       char anchor[MAXPATHLEN])
 {
   bzero(&pfd->pft, sizeof(pfd->pft));
   bzero(&pfd->pfte, sizeof(pfd->pfte));
@@ -58,6 +89,9 @@ static int pf_prepare (struct pf_data *pfd, uint32_t rulenum)
   snprintf(pfd->pfte.anchor, MAXPATHLEN, "%s/%d.%d",
            PF_ANCHOR_NAME, getpid(), rulenum);
 
+  if (anchor != NULL)
+    strncpy(anchor, pfd->pfte.anchor, sizeof(MAXPATHLEN));
+
   return (ioctl(pfd->fd, DIOCXBEGIN, &pfd->pft) == -1 ? -1 : 0);
 }
 
@@ -72,9 +106,8 @@ int fw_add_rule (void *data, uint32_t rulenum, struct rule *rule)
   struct pf_data *pfd = (struct pf_data*) data;
   char anchor[MAXPATHLEN];
 
-  if (pf_prepare(pfd, rulenum) == -1) return -1;
-
-  snprintf(anchor, MAXPATHLEN, "%s/%d.%d", PF_ANCHOR_NAME, getpid(), rulenum);
+  if (pf_prepare(pfd, rulenum, anchor) == -1)
+    return -1;
 
   bzero(&pfd->pfp, sizeof(pfd->pfp));
   strlcpy(pfd->pfp.anchor, anchor, MAXPATHLEN);
@@ -91,8 +124,8 @@ int fw_add_rule (void *data, uint32_t rulenum, struct rule *rule)
   pfd->pfr.rule.rtableid = -1;
   pfd->pfr.rule.quick = 1;
   pfd->pfr.rule.log = 0;
-  if (rule->queue != NULL)
-    strlcpy(pfd->pfr.rule.qname, rule->queue, sizeof(pfd->pfr.rule.qname));
+  if (rule->pf_queue != NULL)
+    strlcpy(pfd->pfr.rule.qname, rule->pf_queue, sizeof(pfd->pfr.rule.qname));
   pfd->pfr.rule.keep_state = 1;
   pfd->pfr.rule.flags = TH_SYN;
   pfd->pfr.rule.flagset = TH_SYN|TH_ACK|TH_FIN|TH_RST;
@@ -127,8 +160,11 @@ int fw_add_rule (void *data, uint32_t rulenum, struct rule *rule)
     memset(&pfd->pfr.rule.dst.addr.v.a.mask.addr8, 255, 4);
   }
 
-  if (ioctl(pfd->fd, DIOCADDRULE, &pfd->pfr) == -1) goto FW_ADD_RULE_FAIL;
-  if (pf_commit(pfd) == -1) goto FW_ADD_RULE_FAIL;
+  if (ioctl(pfd->fd, DIOCADDRULE, &pfd->pfr) == -1)
+    goto FW_ADD_RULE_FAIL;
+
+  if (pf_commit(pfd) == -1)
+    goto FW_ADD_RULE_FAIL;
 
   return 0;
 
@@ -136,6 +172,7 @@ FW_ADD_RULE_FAIL:
   save_errno = errno;
   ioctl(pfd->fd, DIOCXROLLBACK, &pfd->pft);
   errno = save_errno;
+
   return -1;
 }
 
@@ -144,6 +181,6 @@ int fw_del_rule (void *data, uint32_t rulenum)
   struct pf_data *pfd = (struct pf_data*) data;
   int ret;
 
-  ret = pf_prepare(pfd, rulenum);
+  ret = pf_prepare(pfd, rulenum, NULL);
   return (pf_commit(pfd) || ret);
 }
